/* ====================================================================
 * Copyright (c) 1995-1997 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * http_auth_tacacs: authentication via TACACS server
 *
 * rv@tlg.khv.ru (Roman Volkoff)
 */

/*
 * <directory /web/docs/private>
 * Auth_TacacsHost localhost
 * or
 * Auth_Tacacs_Host 194.85.113.20
 *
 * Auth_Tacacs_Key tacacs+_key
 *
 * Auth_Tacacs_Authoritative        on
 *
 * AuthName             example tacacs realm
 * AuthType             basic
 *
 *                      Normal apache/ncsa tokens for access control
 *
 * <limit get post head>
 *   order deny,allow
 *   allow from all
 *
 *   require valid-user
 *                     'valid-user'; allow in any user which has a valid uid/pa>
 *                     pair in the above pwd_table.
 * or
 *   require user smith jones
 *                    Limit access to users who have a valid uid/passwd pair in>
 *                    above pwd_table AND whose uid is 'smith' or 'jones'. Do n>
 *                    the uid's are separated by 'spaces' for historic (ncsa) r>
 *                    So allowing uids with spaces might cause problems.
 *
 *   </limit>
 * </directory>
 *
 */


#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_conf_globals.h"

#include "tac_plus.h"
#include "md5.h"

typedef struct auth_config_struct {
    char *tac_host;
    char *tac_key;
    int   tac_authoritative;
} tac_config_rec;

static void *create_tac_dir_config (pool *p, char *d)
{
    tac_config_rec *sec =
	(tac_config_rec *) pcalloc (p, sizeof(tac_config_rec));
    sec->tac_host = NULL;     /* no tacacs host by default */
    sec->tac_key = NULL;      /* no tacacs key by default */
    sec->tac_authoritative = 1; /* keep the fortress secure by default */
    return sec;
}

static const char *set_tac_slot (cmd_parms *cmd, char *struct_ptr, char *arg)
{
   int offset = (int)cmd->info;
   *(char **)(struct_ptr + offset) = pstrdup (cmd->pool, arg);
   return NULL;
}

static command_rec tac_cmds[] = {
{ "Auth_Tacacs_Host", set_tac_slot,
  (void*)XtOffsetOf(tac_config_rec,tac_host), OR_AUTHCFG, TAKE1,
     "Tacacs-server host. " },
{ "Auth_Tacacs_Key", set_tac_slot,
  (void*)XtOffsetOf(tac_config_rec,tac_key), OR_AUTHCFG, TAKE1,
     "Tacacs-server key. " },
{ "Auth_Tacacs_Authoritative", set_flag_slot,
  (void*)XtOffsetOf(tac_config_rec,tac_authoritative),
    OR_AUTHCFG, FLAG, 
   "Set to 'no' to allow access control to be passed along to lower modules if the UserID is not known to this module" },
{ NULL }
};

module tacacs_module;


/* These functions return 0 if client is OK, and proper error status
 * if not... either AUTH_REQUIRED, if we made a check, and it failed, or
 * SERVER_ERROR, if things are so totally confused that we couldn't
 * figure out how to tell if the client is authorized or not.
 *
 * If they return DECLINED, and all other modules also decline, that's
 * treated by the server core as a configuration error, logged and
 * reported as such.
 */

/* Determine user ID, and check if it really is that user, for HTTP
 * basic authentication...
 */

static int tac_authenticate_basic_user (request_rec *r)
{
    tac_config_rec *sec =
      (tac_config_rec *)get_module_config (r->per_dir_config, &tacacs_module);
    conn_rec *c = r->connection;
    int res,i;
    char             username[256];
    int              result, ok = 0;
    char *sent_pw, *real_pw;
    char *host = sec->tac_host;
    char *key = sec->tac_key;
   char serv_msg[256];
   char data_msg[256];
    
    if ((res = get_basic_auth_pw (r, &sent_pw))) return res;
    
    if(!sec->tac_host)   /* if no tacacs-server host */
        return DECLINED;

    /*
     * Fill in User-Name
     */
    memset (username, 0, sizeof(username));
    strncpy(username, c->user, sizeof(username)-1);
    /*
     * Check in User-Name
     */
    /* send authen */
    tac_connect(host,4,key,49);
    send_authen_start("web",username,2,"");
    i=get_reply_packet(serv_msg,data_msg);
    send_authen_cont(sent_pw,"");
    i=get_reply_packet(serv_msg,data_msg);
    if (i==1) return (OK);
    else return AUTH_REQUIRED;
}
    
/* Checking ID */
    
static int tac_check_user_access (request_rec *r) {
    tac_config_rec *sec =
     (tac_config_rec *)get_module_config (r->per_dir_config, &tacacs_module);
    char *user = r->connection->user;
    int m = r->method_number;
    int method_restricted = 0;
    register int x;
    const char *t, *w;
    array_header *reqs_arr = requires (r);
    require_line *reqs;

    if (!reqs_arr)
        return (OK);
    reqs = (require_line *)reqs_arr->elts;

    for(x=0; x < reqs_arr->nelts; x++) {
      
	if (! (reqs[x].method_mask & (1 << m))) continue;
	
	method_restricted = 1;

        t = reqs[x].requirement;
        w = getword(r->pool, &t, ' ');
        if(!strcmp(w,"valid-user"))
            return OK;
        if(!strcmp(w,"user")) {
            while(t[0]) {
                w = getword_conf (r->pool, &t);
                if(!strcmp(user,w))
                    return OK;
            }
        }
	else if(!strcmp(w,"group"))
	        return DECLINED;	/* DBM group?  Something else? */
    }
    
    if (!method_restricted)
      return OK;

    if (!(sec->tac_authoritative))
      return DECLINED;

    note_basic_auth_failure (r);
    return AUTH_REQUIRED;
}

module tacacs_module = {
   STANDARD_MODULE_STUFF,
   NULL,                        /* initializer */
   create_tac_dir_config,       /* dir config creater */
   NULL,			/* dir merger --- default is to override */
   NULL,			/* server config */
   NULL,			/* merge server config */
   tac_cmds,                    /* command table */
   NULL,			/* handlers */
   NULL,			/* filename translation */
   tac_authenticate_basic_user, /* check_user_id */
   tac_check_user_access,       /* check auth */
   NULL,			/* check access */
   NULL,			/* type_checker */
   NULL,			/* fixups */
   NULL,			/* logger */
   NULL				/* header parser */
};
