/* ====================================================================
 * Copyright (c) 1995 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

/*
 * mod_log_dir.c
 *
 * By Lou Langholtz <http://www.eng.utah.edu/~ldl/>
 *
 * $Id: mod_log_dir.c,v 1.13 1997/07/01 22:31:45 sysldl Exp $
 *
 * This module implements per-directory logging (like the config log
 * module), except that log files must already exist and be writable
 * by the web server.
 *
 * The syntax of the argument to DirLogFormat is the same as for
 * LogFormat from mod_log_config.c, aka config_log_module.
 * The default DirLogFormat also reproduces CLF; see below.
 */

#define DEFAULT_DIR_LOG_FORMAT "%h %l %u %t \"%r\" %s %b"

#include "httpd.h"
#include "http_config.h"
#include "http_core.h" /* For REMOTE_NAME */

#define DLS_ALLOW_UNSET -1
#define DLS_ALLOW_FALSE 0
#define DLS_ALLOW_TRUE 1

module dir_log_module;

/*
 * Don't O_CREAT any files for reasons of security.
 *
 * It's much safer if we require that the file already exist,
 * be writable by the web server, and be owned by the same uid
 * as owns the .htaccess file that configured us.
 */
static int xfer_flags = ( O_WRONLY | O_APPEND );

typedef struct {
    char *fname;
    array_header *format;
    int log_fd;
    int required_uid;
    int allow;
} dir_log_state;

/*
 * Format items...
 * Copied from mod_log_config.c
 */

typedef char *(*item_key_func)(request_rec *, char *);

typedef struct {
    item_key_func func;
    char *arg;
    int condition_sense;
    int want_orig;
    array_header *conditions;
} log_format_item;

/*****************************************************************
 *
 * Parsing the log format string
 */

#if MODULE_MAGIC_NUMBER < 19961125
extern array_header *parse_log_string(pool *p, char *s, char **err);
#else
extern array_header *parse_log_string(pool *p, const char *s, const char **err);
#endif
extern char *process_item(request_rec *r, request_rec *orig,
	log_format_item *item);

/*****************************************************************
 *
 * Actually logging.
 *
 * Note that at the point of a logger, it's too late to return anything
 * but OK or DECLINED since the page has already been served up.
 * So we must run a access checker (ie. dir_log_access()) to be able
 * to return any indication at all to the browser that there's an error.
 */

int dir_log_access(request_rec *r)
{
    dir_log_state *dls;
    struct stat statb;

    dls = ap_get_module_config (r->per_dir_config, &dir_log_module);
    if (!dls || dls->fname == NULL)
	return DECLINED;

    /*
     * Since we can't register our own override options handle
     * the case where someone's tried to override logging even
     * though that's not allowed. This way they at least get some
     * feedback that something's wrong instead of just silently
     * not getting anything logged.
     */
    if (dls->allow != DLS_ALLOW_TRUE) {
	/*
	 * Try and make a reasonable guess at what our actual
	 * error log message would be if we actually could use
	 * overrides.
	 */
    	if (dls->fname != NULL) {
	    ap_log_reason(
		ap_pstrcat(r->pool, "DirTransferLog", " not allowed here", NULL),
		ap_pstrcat(r->pool, r->htaccess->dir, ".htaccess", NULL),
		r);
	    return SERVER_ERROR;
	}
	if (dls->format != NULL) {
	    ap_log_reason(
		ap_pstrcat(r->pool, "DirLogFormat", " not allowed here", NULL),
		ap_pstrcat(r->pool, r->htaccess->dir, ".htaccess", NULL),
		r);
	    return SERVER_ERROR;
	}
	return DECLINED;
    }

    if (dls->log_fd < 0) {
	dls->log_fd = open(dls->fname, xfer_flags);
	if (dls->log_fd < 0) {
	    ap_log_reason(ap_pstrcat(r->pool, "can't open logfile ",
		dls->fname, NULL), r->filename, r);
	    return SERVER_ERROR;
	}
    }

    if (fstat(dls->log_fd, &statb) < 0) {
	close(dls->log_fd);
	dls->log_fd = -1;
	ap_log_reason(ap_pstrcat(r->pool, "can't fstat logfile ",
	    dls->fname, NULL), r->filename, r);
	return SERVER_ERROR;
    }

    if (dls->required_uid != statb.st_uid) {
	char id1[10], id2[10];

	close(dls->log_fd);
	dls->log_fd = -1;
	sprintf(id1, "%d", dls->required_uid);
	sprintf(id2, "%d", statb.st_uid);
	ap_log_reason(ap_pstrcat(r->pool,
	    "logfile ", dls->fname, " owned by uid ", id2,
	    " instead of ", id1, NULL), r->filename, r);
	return SERVER_ERROR;
    }

    return OK;
}

int dir_log_transaction(request_rec *r)
{
    dir_log_state *dls;
    array_header *strsa;
    log_format_item *items;
    char *str, **strs, *s;
    request_rec *orig;
    int i;
    int len = 0;

    dls = ap_get_module_config (r->per_dir_config, &dir_log_module);
    if (!dls || dls->allow != DLS_ALLOW_TRUE || dls->fname == NULL
    || dls->log_fd < 0)
	return DECLINED;

    if (dls->format == NULL) {
#if MODULE_MAGIC_NUMBER < 19961125
	char *dummy = NULL;
#else
	const char *dummy = NULL;
#endif
	dls->format = parse_log_string(r->pool,
	    DEFAULT_DIR_LOG_FORMAT, &dummy);
    }
    strsa= ap_make_array(r->pool, dls->format->nelts,sizeof(char*));
    items = (log_format_item *)dls->format->elts;

    orig = r;
    while (orig->prev) orig = orig->prev;
    while (r->next) r = r->next;

    for (i = 0; i < dls->format->nelts; ++i)
	*((char**)ap_push_array (strsa)) = process_item (r, orig, &items[i]);

    strs = (char **)strsa->elts;
    
    for (i = 0; i < dls->format->nelts; ++i)
	len += strlen (strs[i]);

    str = ap_palloc (r->pool, len + 1);

    for (i = 0, s = str; i < dls->format->nelts; ++i) {
	strcpy (s, strs[i]);
	s += strlen (strs[i]);
    }

    write(dls->log_fd, str, strlen(str));
    close(dls->log_fd);
    dls->log_fd = -1;

    return OK;
}

/*****************************************************************
 *
 * Module glue...
 */

void *make_dir_log_state (pool *p, char *dummy)
{
    dir_log_state *dls =
      (dir_log_state *)ap_palloc (p, sizeof (dir_log_state));

    dls->fname = NULL;
    dls->format = NULL;
    dls->log_fd = -1;
    dls->required_uid = -1;
    dls->allow = DLS_ALLOW_UNSET;

    return (void *)dls;
}

#ifdef notdef
void dump_dir_log_state(dir_log_state *dls)
{
    fprintf(stderr, "%s %d %d %d\n",
	dls->fname? dls->fname: "null",
	dls->log_fd, dls->required_uid, dls->allow);
}
#endif

void *merge_dir_log_state(pool *p, void *parent, void *sub)
{
    dir_log_state *par = (dir_log_state *)parent;
    dir_log_state *chld = (dir_log_state *)sub;
    dir_log_state *mrg = (dir_log_state *)ap_palloc(p, sizeof(*mrg));

    mrg->fname = chld->fname;
    mrg->format = chld->format;
    mrg->log_fd = chld->log_fd;
    mrg->required_uid = chld->required_uid;
    mrg->allow = (chld->allow != DLS_ALLOW_UNSET)?
	chld->allow: ((par->allow != DLS_ALLOW_UNSET)?
	    par->allow: DLS_ALLOW_UNSET);
    return mrg;
}

#if MODULE_MAGIC_NUMBER < 19961125
char *set_dir_log_allow(cmd_parms *cmd, void *dv, int arg)
#else
const char *set_dir_log_allow(cmd_parms *cmd, void *dv, int arg)
#endif
{
    /*
     * This really should be dealt with as an override directive
     * so users can get some feedback when they've tried using
     * DirTransferLog even though it's not allowed in that dir.
     * Need access to core_dir_config to add an override bit.
     * This is supposedly coming in the next release of Apache.
     */
    dir_log_state *dls = (dir_log_state *)dv;
    dls->allow = arg? DLS_ALLOW_TRUE: DLS_ALLOW_FALSE;
    return NULL;
}

#if MODULE_MAGIC_NUMBER < 19961125
char *set_dir_log_file (cmd_parms *parms, void *dv, char *arg)
#else
const char *set_dir_log_file (cmd_parms *parms, void *dv, char *arg)
#endif
{
    dir_log_state *dls = (dir_log_state *)dv;
    struct stat statb;

    if (*arg != '/')
	return "DirTransferLog not absolute pathname";

    if (!(parms->path) || stat(parms->path, &statb) < 0)
	return "can't fstat config file";

    dls->fname = arg;
    dls->required_uid = statb.st_uid;

    return NULL;
}

#if MODULE_MAGIC_NUMBER < 19961125
char *dir_log_format (cmd_parms *cmd, void *dv, char *arg)
#else
const char *dir_log_format (cmd_parms *cmd, void *dv, char *arg)
#endif
{
    dir_log_state *dls = (dir_log_state *)dv;
#if MODULE_MAGIC_NUMBER < 19961125
    char *err_string = NULL;
#else
    const char *err_string = NULL;
#endif

    dls->format = parse_log_string (cmd->pool, arg, &err_string);
    return err_string;
}

command_rec dir_log_cmds[] = {
{ "DirLogAllow", set_dir_log_allow, NULL, ACCESS_CONF, FLAG,
    "'on' or 'off' to allow or disallow per-dir logging (default 'off')" },
{ "DirTransferLog", set_dir_log_file, NULL, OR_FILEINFO, TAKE1,
    "absolute path of pre-existing writable access log owned by same user as is *AccessFileName* for directory" },
{ "DirLogFormat", dir_log_format, NULL, OR_FILEINFO, TAKE1,
    "log format string for per-directory access log" },
{ NULL }
};

module dir_log_module = {
   STANDARD_MODULE_STUFF,
   NULL,		/* initializer */
   make_dir_log_state,	/* create per-dir config */
   merge_dir_log_state,		/* dir merger --- default is to override */
   NULL,		/* server config */
   NULL,		/* merge server config */
   dir_log_cmds,	/* command table */
   NULL,		/* handlers */
   NULL,		/* filename translation */
   NULL,		/* check_user_id */
   NULL,		/* check auth */
   dir_log_access,	/* check access */
   NULL,		/* type_checker */
   NULL,		/* fixups */
   dir_log_transaction	/* logger */
};
