
/* ====================================================================
 * Copyright (c) 1996 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission.
 *
 * 5. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */


/*
**  mod_rewrite_compat.c -- Backward-Compatibility Module Code
**                       _                            _ _ 
**   _ __ ___   ___   __| |    _ __ _____      ___ __(_) |_ ___ 
**  | '_ ` _ \ / _ \ / _` |   | '__/ _ \ \ /\ / / '__| | __/ _ \
**  | | | | | | (_) | (_| |   | | |  __/\ V  V /| |  | | ||  __/
**  |_| |_| |_|\___/ \__,_|___|_|  \___| \_/\_/ |_|  |_|\__\___|
**                       |_____|
**
**  URL Rewriting Module, Version 2.3.0 (09-09-1996)
**
**  This module uses a rule-based rewriting engine (based on a
**  regular-expression parser) to rewrite requested URLs on the fly. 
**  
**  It supports an unlimited number of additional rule conditions (which can
**  operate on a lot of variables, even on HTTP headers) for granular
**  matching and even external database lookups (either via plain text
**  tables, DBM hash files or even external processes) for advanced URL
**  substitution.
**  
**  It operates on the full URLs (including the PATH_INFO part) both in
**  per-server context (httpd.conf) and per-dir context (.htaccess) and even
**  can generate QUERY_INFO parts on result.   The rewriting result finally
**  can lead to internal subprocessing, external request redirection or even
**  to internal proxy throughput.
**
**  The documentation and latest release can be found on
**  http://www.engelschall.com/sw/mod_rewrite/
**
**  Copyright (c) 1996 The Apache Group,    All rights reserved.
**  Copyright (c) 1996 Ralf S. Engelschall, All rights reserved.
**
**  Written for The Apache Group by
**      Ralf S. Engelschall
**      rse@engelschall.com
**      http://www.engelschall.com/~rse
*/




    /* from the underlaying Unix system ... */
#include <string.h>
#include <stdarg.h>

    /* from the Apache server ... */
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"

    /* now our own stuff ... */
#include "mod_rewrite.h"




/*
**
**  forward declarations
**
*/

    /* static config */
static command_rec rewrite_compat_cmds[];

    /* internal stuff */
static void dbg(char *str, ...);

    /* the directives, simulating mod_alias */
static char *cmd_alias(cmd_parms *cmd, void *dummy, char *f, char *r);
static char *cmd_scriptalias(cmd_parms *cmd, void *dummy, char *f, char *r);
static char *cmd_redirect(cmd_parms *cmd, void *dummy, char *f, char *url);

    /* the ScriptAlias kludge hook */
static int hook_mimetype(request_rec *r);

    /* the directives, simulating mod_userdir */
static char *cmd_userdir(cmd_parms *cmd, void *dummy, char *arg);




/*
**
**  our interface to the Apache server kernel
**
*/

module rewrite_compat_module = {
   STANDARD_MODULE_STUFF,
   NULL,                        /* initializer */
   NULL,                        /* dir config creater */
   NULL,                        /* dir merger --- default is to override */
   NULL,                        /* server config */
   NULL,                        /* merge server configs */
   rewrite_compat_cmds,         /* command table */
   NULL,                        /* handlers */
   NULL,                        /* filename translation */
   NULL,                        /* check_user_id */
   NULL,                        /* check auth */
   NULL,                        /* check access */
   hook_mimetype,               /* type_checker */
   NULL,                        /* fixups */
   NULL                         /* logger */
};

static command_rec rewrite_compat_cmds[] = {
    /*
     *  fake mod_alias
     */
    { "Alias", cmd_alias, NULL, RSRC_CONF, TAKE2, 
      "a fakename and a realname"},
    { "ScriptAlias", cmd_scriptalias, NULL, RSRC_CONF, TAKE2, 
      "a fakename and a realname"},
    { "Redirect", cmd_redirect, NULL, OR_FILEINFO, TAKE2, 
      "a document to be redirected, then the destination URL" },
    /*
     *  fake mod_userdir
     */
    { "UserDir", cmd_userdir, NULL, RSRC_CONF, RAW_ARGS,
      "the public subdirectory in users' home directories, or 'disabled'" },
    { NULL }
};




/*
**
**  Internal stuff
**
*/

#if DEBUG
static void dbg(char *str, ...)
{
    va_list ap;

    va_start(ap, str);
    vfprintf(stderr, str, ap);
    va_end(ap);
    return;
}
#endif




/*
**
**  fake mod_alias
**
**  Syntax to be replaced:
**    Alias        /abc/def /XYZ
**    ScriptAlias  /abc/def /XYZ                     (**)
**    Redirect     /abc/def http://<host>/XYZ/
**    
**  Replacement Syntax:
**    RewriteRule ^/abc/def(.*) /XZY$1               [L]
**    RewriteRule ^/abc/def(.*) /XZY$1               [T=application/x-httpd-cgi,L] (**)
**    RewriteRule ^/abc/def(.*) http://<host>/XYZ/$1 [R,L]
**
**  (**)The mod_alias had a kludge for an ugly handshake with mod_cgi to
**      automatically map all files in a ScriptAlias directory to MIME-type
**      "application/x-httpd-cgi" and to force mod_cgi implicitly think there
**      is a ExecCGI option for this dir. We have to simulate this via
**      setting the same request note as mod_alias. This is done by reading
**      mod_rewrite's request note after we have used the T=<type> flag!
**
*/

char *cmd_alias(cmd_parms *cmd, void *dummy, char *f, char *r)
{
    char cmdline[MAX_STRING_LEN];

    sprintf(cmdline, "^%s(.*) %s$1 [L]", f, r);
#if DEBUG
    dbg("<< 'Alias %s %s'\n", f, r);
    dbg(">> 'RewriteRule %s'\n", cmdline);
#endif
    return cmd_rewriterule(cmd, dummy, cmdline);
}

char *cmd_scriptalias(cmd_parms *cmd, void *dummy, char *f, char *r)
{
    char cmdline[MAX_STRING_LEN];

    sprintf(cmdline, "^%s(.*) %s$1 [T=%s,L]", f, r, CGI_MAGIC_TYPE);
#if DEBUG
    dbg("<< 'ScriptAlias %s %s'\n", f, r);
    dbg(">> 'RewriteRule %s'\n", cmdline);
#endif
    return cmd_rewriterule(cmd, dummy, cmdline);
}

char *cmd_redirect(cmd_parms *cmd, void *dummy, char *f, char *url)
{
    char cmdline[MAX_STRING_LEN];

    if (!is_url(url)) 
        return "Redirect to non-URL";
    sprintf(cmdline, "^%s(.*) %s$1 [R,L]", f, url);
#if DEBUG
    dbg("<< 'Redirect %s %s'\n", f, url);
    dbg(">> 'RewriteRule %s'\n", cmdline);
#endif
    return cmd_rewriterule(cmd, dummy, cmdline);
}


/* this is the internal note mod_rewrite uses */
#define REWRITE_FORCED_MIMETYPE_NOTEVAR "rewrite-forced-mimetype"
/* this is the internal note mod_alias used for handshake with mod_cgi */
#define ALIAS_FORCED_MIMETYPE_NOTEVAR "alias-forced-type"

int hook_mimetype(request_rec *r)
{
    char *t;
    
    /* now check if we have to force a MIME-type */
    t = table_get(r->notes, REWRITE_FORCED_MIMETYPE_NOTEVAR);
    if (t != NULL) {
        if (strcmp(t, CGI_MAGIC_TYPE) == 0) {
            table_set(r->notes, ALIAS_FORCED_MIMETYPE_NOTEVAR, t);
        }
    }
    return DECLINED;
}



/*
**
**  fake mod_userdir
**
**  Syntax to be replaced:
**    [1]  UserDir public_html     
**    [2]  UserDir /usr/web
**    [3]  UserDir /home/ * /www
**    [4]  UserDir http://x/users
**    [5]  UserDir http://x/ * /y
**    
**  Replacement Syntax:
**    RewriteRule ^/~([a-z]+)(.*)  ~$1/public_html$2
**    RewriteRule ^/~([a-z]+.*)    /usr/web/$1
**    RewriteRule ^/~([a-z]+)(.*)  /home/$1/www$2
**    RewriteRule ^/~([a-z]+.*)    http://x/users/$1
**    RewriteRule ^/~([a-z]+)(.*)  http://x/$1/y$2
**
**  Be careful: When you substitute mod_userdir.c by this module
**              you have to make sure that there is at least one
**              "UserDir" directive in your config, because
**              this module cannot set a default rule.
**              In other words: If you use this module and don't
**              use the "UserDir" directiuve than you cannot
**              access userdirs as long as you don't add an own
**              explicit "RewriteRule" directive to do a similar job.
**        
*/

char *cmd_userdir(cmd_parms *cmd, void *dummy, char *arg)
{
    char *cp;
    char *cp2;
    char *pat;
    char cmdline[MAX_STRING_LEN];

    if (!strcasecmp(arg, "disabled"))
        return NULL;

#ifdef DEBUG
    dbg("<< 'UserDir %s'\n", arg);
#endif
    for (cp = arg; *cp != '\0'; ) {
        /* strip whitespaces */
        for ( ; *cp != '\0' && (*cp == ' ' || *cp == '\t'); cp++)
            ;
        if (*cp == '\0') 
            break;

        pat = cp;
        /* read up to whitespaces or end of string */
        for ( ; *cp != '\0' && (*cp != ' ' && *cp != '\t'); cp++)
            ;
        *cp++ = '\0';

        if (pat[0] != '/' && strncmp(pat, "http://", 7) != 0) {
            /* [1] */
            sprintf(cmdline, "^/~([a-z]+)(.*) %s$1$2 [L]", pat);
        }
        else if (pat[0] == '/' || strncmp(pat, "http://", 7) == 0) {
            if ((cp2 = strchr(pat, '*')) == NULL) {
                /* [2] and [4] */
                sprintf(cmdline, "^/~([a-z]+.*) %s$1 [L]", pat);
            }
            else {
                /* [3] and [5] */
                *cp2++ = '\0';    
                sprintf(cmdline, "^/~([a-z]+)(.*) %s$1%s$2 [L]", pat, cp2);
            }
        }
        else {
            fprintf(stderr, "UserDir: bad argument '%s'\n", pat);
            exit(1);
        }
#ifdef DEBUG
        dbg(">> 'RewriteRule %s'\n", cmdline);
#endif
        (void)cmd_rewriterule(cmd, dummy, cmdline);
    }
    return NULL;
}


/*EOF*/
